@@ -4,24 +4,23 @@ import time
 from collections import namedtuple
 from datetime import timedelta
 from distutils.version import StrictVersion
-from os import chdir, get_terminal_size, path
-from pathlib import Path
+from os import get_terminal_size
 from queue import Empty
-from random import choice, randint, sample
+from random import randint
 
-import requests
 import soco
 import soco.alarms
 import tabulate
-import xmltodict
+from soco.exceptions import NotSupportedException
 
-from .m3u_parser import parse_m3u
-from .play_local_file import is_supported_type, play_local_file
+from .play_local_file import play_local_file
+from .play_m3u_file import play_m3u_file
 from .speaker_info import print_speaker_table
 from .utils import (
     convert_to_seconds,
     convert_true_false,
     error_and_exit,
+    event_unsubscribe,
     get_right_hand_speaker,
     get_speaker,
     one_or_more_parameters,
@@ -1210,7 +1209,32 @@ def wait_stop(speaker, action, args, soco_function, use_local_speaker_list):
                         speaker.player_name, event.variables["transport_state"]
                     )
                 )
-                sub.unsubscribe()
+                event_unsubscribe(sub)
+                return True
+        except Empty:
+            pass
+
+
+@zero_parameters
+def wait_stop_not_pause(speaker, action, args, soco_function, use_local_speaker_list):
+    try:
+        sub = speaker.avTransport.subscribe(auto_renew=True)
+    except Exception as e:
+        error_and_exit("Exception {}".format(e))
+    while True:
+        try:
+            event = sub.events.get(timeout=1.0)
+            if event.variables["transport_state"] not in [
+                "PLAYING",
+                "TRANSITIONING",
+                "PAUSED_PLAYBACK",
+            ]:
+                logging.info(
+                    "Speaker '{}' in state '{}'".format(
+                        speaker.player_name, event.variables["transport_state"]
+                    )
+                )
+                event_unsubscribe(sub)
                 return True
         except Empty:
             pass
@@ -1241,7 +1265,7 @@ def wait_stopped_for(speaker, action, args, soco_function, use_local_speaker_lis
             )
             if event.variables["transport_state"] not in playing_states:
                 logging.info("Speaker is not in states {}".format(playing_states))
-                sub.unsubscribe()
+                event_unsubscribe(sub)
                 # ToDo: Should really return here and do this some other way ...
                 #       this is what's requiring the SIGKILL
 
@@ -1303,7 +1327,7 @@ def wait_start(speaker, action, args, soco_function, use_local_speaker_list):
                         speaker.player_name, event.variables["transport_state"]
                     )
                 )
-                sub.unsubscribe()
+                event_unsubscribe(sub)
                 return True
         except Empty:
             pass
@@ -1645,38 +1669,21 @@ def cue_favourite_radio_station(
 
 @zero_parameters
 def battery(speaker, action, args, soco_function, use_local_speaker_list):
-    info = speaker.get_speaker_info()
-    logging.info("Retrieved speaker info: {}".format(info))
-    if info["model_name"] not in ["Sonos Move"]:
+    try:
+        battery_status = speaker.get_battery_info()
+    except NotSupportedException:
         error_and_exit(
-            "Speaker '{}' doesn't have a battery".format(speaker.player_name)
+            "Battery status not supported by '{}'".format(speaker.player_name)
         )
         return False
-
-    # Retrieve information from the speaker's support URL
-    url = "http://" + speaker.ip_address + ":1400/support/review"
-    try:
-        response = requests.get(url)
     except:
-        error_and_exit("Failed to retrieve speaker information")
-        return False
-    if response.status_code != 200:
-        error_and_exit("Failed to retrieve speaker information")
+        error_and_exit("Unable to retrieve battery status")
         return False
 
-    # Traverse XML to obtain the battery information
-    try:
-        data = xmltodict.parse(response.text)
-        zp_list = data["ZPNetworkInfo"]["ZPSupportInfo"]
-        for zp in zp_list:
-            if zp["ZPInfo"]["IPAddress"] == speaker.ip_address:
-                battery_status = zp["LocalBatteryStatus"]["Data"]
-                for item in battery_status:
-                    print("  {}: {}".format(item["@name"], item["#text"]))
-                return True
-    except:
-        error_and_exit("Error in the information returned by the speaker")
-        return False
+    for key, value in battery_status.items():
+        print("  " + key + ": " + str(value))
+
+    return True
 
 
 @one_parameter
@@ -1724,9 +1731,13 @@ def add_uri_to_queue(speaker, action, args, soco_function, use_local_speaker_lis
     return True
 
 
-@one_parameter
+@one_or_more_parameters
 def play_file(speaker, action, args, soco_function, use_local_speaker_list):
-    return play_local_file(speaker, args[0])
+    for audio_file in args:
+        result = play_local_file(speaker, audio_file)
+        if not result:
+            return False
+    return True
 
 
 @one_or_two_parameters
@@ -1735,65 +1746,75 @@ def play_m3u(speaker, action, args, soco_function, use_local_speaker_list):
     options = "" if len(args) == 1 else args[1]
     options = options.lower()
 
-    # Check for invalid options
-    invalid = set(options) - set("psr")
-    if invalid:
-        error_and_exit("Invalid option(s) '{}' supplied".format(invalid))
-        return False
+    play_m3u_file(speaker, m3u_file, options=options)
+    return True
 
-    if not (m3u_file.lower().endswith(".m3u") or m3u_file.lower().endswith(".m3u8")):
-        error_and_exit(
-            "Filename '{}' does not end in '.m3u' or '.m3u8'".format(m3u_file)
-        )
-        return False
 
-    if not path.exists(m3u_file):
-        error_and_exit("File '{}' not found".format(m3u_file))
-        return False
+@zero_or_one_parameter
+def buttons(speaker, action, args, soco_function, use_local_speaker_list):
+    """Enable or disable a speaker's buttons"""
+    np = len(args)
+    if np == 0:
+        state = "on" if speaker.buttons_enabled else "off"
+        print(state)
+    elif np == 1:
+        arg = args[0].lower()
+        if arg == "on":
+            speaker.buttons_enabled = True
+        elif arg == "off":
+            speaker.buttons_enabled = False
+        else:
+            parameter_type_error(action, "on|off")
+    return True
 
-    logging.info("Parsing M3U file '{}'".format(m3u_file))
-    tracks = parse_m3u(m3u_file)
-    if not tracks:
-        error_and_exit("No tracks found in '{}'".format(m3u_file))
 
-    logging.info("Found {} tracks".format(len(tracks)))
-
-    if "r" in options:
-        # Choose a single random track
-        track = choice(tracks)
-        tracks = [track]
-        logging.info("Choosing random track: {}".format(track.path))
-
-    elif "s" in options:
-        logging.info("Shuffling playlist")
-        # For some reason, 'shuffle(tracks)' does not work
-        tracks = sample(tracks, len(tracks))
-
-    directory, _ = path.split(m3u_file)
-    if directory != "":
-        chdir(directory)
-
-    pad = len(str(len(tracks)))
-    for index, track in enumerate(tracks):
-        abs_filename = str(Path(track.path).absolute())
-        logging.info("Convert '{}' to '{}'".format(track.path, abs_filename))
-
-        if not path.exists(abs_filename):
-            print("Error: file not found:", abs_filename)
-            continue
-
-        if not is_supported_type(abs_filename):
-            print("Error: unsupported file type:", abs_filename)
-            continue
-
-        if "p" in options:
-            print(
-                "Playing {} of {}:".format(str(index + 1).zfill(pad), len(tracks)),
-                abs_filename,
+@zero_or_one_parameter
+def fixed_volume(speaker, action, args, soco_function, use_local_speaker_list):
+    """Enable or disable whether a Connect or Port has its Fixed Volume set"""
+    np = len(args)
+    if np == 0:
+        state = "on" if speaker.fixed_volume else "off"
+        print(state)
+    elif np == 1:
+        arg = args[0].lower()
+        try:
+            if arg == "on":
+                speaker.fixed_volume = True
+            elif arg == "off":
+                speaker.fixed_volume = False
+            else:
+                parameter_type_error(action, "on|off")
+        except:
+            error_and_exit(
+                "Fixed Volume feature not supported by '{}'".format(speaker.player_name)
             )
+            return False
+    return True
 
-        play_local_file(speaker, abs_filename)
 
+@zero_or_one_parameter
+def trueplay(speaker, action, args, soco_function, use_local_speaker_list):
+    """Enable or disable whether a Trueplay profile is enabled"""
+    np = len(args)
+    if np == 0:
+        state = "on" if speaker.trueplay else "off"
+        print(state)
+    elif np == 1:
+        arg = args[0].lower()
+        try:
+            if arg == "on":
+                speaker.trueplay = True
+            elif arg == "off":
+                speaker.trueplay = False
+            else:
+                parameter_type_error(action, "on|off")
+        except:
+            error_and_exit(
+                "No Trueplay profile available for '{}' (or Trueplay not supported)".format(
+                    speaker.player_name
+                )
+            )
+            return False
     return True
 
 
@@ -2071,4 +2092,9 @@ actions = {
     "play_m3u": SonosFunction(play_m3u, ""),
     "play_local_m3u": SonosFunction(play_m3u, ""),
     "add_uri_to_queue": SonosFunction(add_uri_to_queue, ""),
+    "wait_stop_not_pause": SonosFunction(wait_stop_not_pause, ""),
+    "wsnp": SonosFunction(wait_stop_not_pause, ""),
+    "buttons": SonosFunction(buttons, ""),
+    "fixed_volume": SonosFunction(fixed_volume, ""),
+    "trueplay": SonosFunction(trueplay, ""),
 }
