@@ -4,14 +4,29 @@ import os
 import pickle
 import sys
 from collections.abc import Sequence
+from platform import python_version
 from signal import SIGTERM
+from time import sleep
 
 import soco
 
 from .__init__ import __version__
+from .match_speaker_names import speaker_name_matches
 from .speakers import Speakers
 
 
+def event_unsubscribe(sub):
+    # Insert a brief pause before event unsubscription prevents lockups,
+    # by yielding the thread
+    pause_seconds = 0.2
+    logging.info(
+        "Unsubscribing from events ... yield by pausing for {}s".format(pause_seconds)
+    )
+    sleep(pause_seconds)
+    sub.unsubscribe()
+    logging.info("Unsubscribed")
+
+
 # Error handling
 def error_and_exit(msg):
     # Print to stderr
@@ -185,6 +200,7 @@ def convert_true_false(true_or_false, conversion="YesOrNo"):
 def version():
     print("soco-cli version: {}".format(__version__))
     print("soco version:     {}".format(soco.__version__))
+    print("python version:   {}".format(python_version()))
 
 
 def docs():
@@ -304,17 +320,123 @@ def set_speaker_list(s):
     speaker_list = s
 
 
+SCAN_TIMEOUT = 0.1
+
+
+class SpeakerCache:
+    def __init__(self):
+        # _cache contains (soco_instance, speaker_name) tuples
+        self._cache = set()
+
+    @property
+    def exists(self):
+        return True if self._cache else False
+
+    def update(self):
+        speakers = soco.discovery.scan_network(
+            multi_household=True, scan_timeout=SCAN_TIMEOUT
+        )
+        if speakers:
+            logging.info("Adding speakers to cache")
+            for speaker in speakers:
+                self._cache.add((speaker, speaker.player_name))
+        else:
+            logging.info("No speakers found to cache")
+        return None
+
+    def add(self, speaker):
+        logging.info("Adding speaker to cache")
+        self._cache.add((speaker, speaker.player_name))
+
+    def find_indirect(self, name):
+        speakers_found = set()
+        speakers_found_names = set()
+        for cached, cached_name in self._cache:
+            for speaker in cached.visible_zones:
+                match, exact = speaker_name_matches(name, speaker.player_name)
+                if match and exact:
+                    return speaker
+                if match and not exact:
+                    speakers_found.add(speaker)
+                    speakers_found_names.add(speaker.player_name)
+
+        if len(speakers_found) == 1:
+            return speakers_found.pop()
+
+        elif len(speakers_found) > 1:
+            error_and_exit("'{}' is ambiguous: {}".format(name, speakers_found_names))
+
+        else:
+            return None
+
+    def find(self, name):
+        speakers_found = set()
+        speakers_found_names = set()
+        for speaker, speaker_name in self._cache:
+            match, exact = speaker_name_matches(name, speaker_name)
+            if match and exact:
+                return speaker
+            if match and not exact:
+                speakers_found.add(speaker)
+                speakers_found_names.add(speaker_name)
+
+        if len(speakers_found) == 1:
+            return speakers_found.pop()
+
+        elif len(speakers_found) > 1:
+            error_and_exit(
+                "Speaker name '{}' is ambiguous within {}".format(
+                    name, speakers_found_names
+                )
+            )
+
+        else:
+            return None
+
+
+# Single instance of the speaker cache
+cache = SpeakerCache()
+
+
 def get_speaker(name, local=False):
-    # Allow the use of an IP address even if 'local' is specified
+    # Use an IP address
+    # (Allow the use of an IP address even if 'local' is specified)
     if Speakers.is_ipv4_address(name):
         logging.info("Using IP address instead of speaker name")
         return soco.SoCo(name)
+
+    # Use the local speaker list
     if local:
         logging.info("Using local speaker list")
         return speaker_list.find(name)
+
+    # Use discovery
     else:
-        logging.info("Using SoCo speaker discovery")
-        return soco.discovery.by_name(name)
+        # Try various lookup methods in order of expense,
+        # and cache results where possible
+        speaker = None
+        if not speaker:
+            logging.info("Trying direct cache lookup")
+            speaker = cache.find(name)
+        if not speaker:
+            logging.info("Trying indirect cache lookup")
+            speaker = cache.find_indirect(name)
+        if not speaker:
+            logging.info("Trying discovery_by_name with network scan fallback")
+            speaker = soco.discovery.by_name(
+                name, allow_network_scan=True, scan_timeout=SCAN_TIMEOUT
+            )
+            if speaker:
+                cache.add(speaker)
+        if not speaker:
+            logging.info("Trying full multi-household network scan")
+            cache.update()
+            speaker = cache.find(name)
+        if speaker:
+            logging.info("Successful speaker discovery")
+        else:
+            logging.info("Failed to discover speaker")
+        return speaker
 
 
 def get_right_hand_speaker(left_hand_speaker):
